# coding: utf-8
from collections import defaultdict
from datetime import datetime
from datetime import time
from itertools import groupby

from django.contrib.contenttypes.models import ContentType
from django.db.models import Count
from edu_to_analytics.service.types import Direction
from edu_to_analytics.service.types import Item
from edu_to_analytics.service.types import Lesson
from edu_to_analytics.service.types import Mark
from edu_to_analytics.service.types import Root
from edu_to_analytics.service.types import UnitsRoot
from educommon.m3 import get_id_value
from educommon.utils.misc import cached_property
from lxml.etree import tostring
from m3.actions.exceptions import ApplicationLogicException
from m3_ext.ui.fields.simple import ExtCheckBox
from spyne.util.xml import get_object_as_xml

from extedu.auth.esia_saml.events import AUTH_LOGIN_ESIA
from extedu.core.dict.models import UnitType
from extedu.core.period.models import Period
from extedu.core.utils.db import get_model_instance
from extedu.core.utils.misc import extract_bool
from extedu.direction.models import Direction as DirectionModel
from extedu.employee.models import Employee
from extedu.event_log.events import AUTH_LOGIN
from extedu.event_log.events import MODULE_AUTH
from extedu.event_log.models import EventLog
from extedu.group.models import Group as GroupModel
from extedu.group.models import SubGroupPupil
from extedu.lesson.models import Lesson as LessonModel
from extedu.lesson.models import Attendance
from extedu.lesson.models import NextLessonHomework
from extedu.mark.models import Mark as MarkModel
from extedu.mark.utils import normalize_mark
from extedu.pupil.models import Pupil as PupilModel
from extedu.unit.models import Unit
from six.moves import range

from .models import SyncObject


def _get_unit_administration(unit):
    if unit.is_administration:
        return unit
    else:
        query = unit.get_ancestors().filter(
            type=UnitType.ADMINISTRATION
        ).order_by('-level')

        return query[0] if query.exists() else None


def _build_record(user_id, day, records):
    first_record = records[0]

    result = Item()

    result.count = len(records)
    result.teacher_id = user_id
    result.fullname = first_record.user_credentials
    result.date = day.strftime('%d.%m.%Y')
    result.times = ', '.join(r.datetime.strftime('%H:%M') for r in records)

    unit = get_model_instance(Unit, first_record.unit_id)
    if unit:
        result.school_id = first_record.unit_id
        result.school_shortname = unit.short_name
        result.school_name = unit.full_name

        administration = _get_unit_administration(unit)
        if administration:
            result.control_unit = administration.short_name
            result.control_unit_id = administration.id
        else:
            result.control_unit = ''
            result.control_unit_id = 0
    else:
        result.control_unit = ''
        result.control_unit_id = 0
        result.school_id = 0
        result.school_shortname = first_record.unit_shortname
        result.school_name = first_record.unit_shortname

    return result


def _get_data(date_from, date_to, event_type):
    """Возвращает данные о количестве аутентификаций сотрудников.

    :param date_from: Дата начала периода, за который нужно загрузить
        информацию.
    :type date_from: datetime.date

    :param date_to: Дата окончания периода, за который нужно загрузить
        информацию.
    :type date_to: datetime.date

    :param str event_type: Тип события в журнале событий (``AUTH_LOGIN`` или
        ``AUTH_ESIA_LOGIN``).

    :rtype: :class:`edu_to_analytics.service.types.Root`
    """
    datetime_range = (
        datetime.combine(date_from, time()),
        datetime.combine(date_to, time(23, 59, 59, 999999)),
    )

    employee_users = Employee.objects.values_list(
        'person__user', flat=True
    ).distinct()

    log_records = EventLog.objects.filter(
        user_id__in=tuple(employee_users),
        module=MODULE_AUTH,
        event=event_type,
        datetime__range=datetime_range,
    ).order_by('user_id', 'datetime')

    employees = []
    for (user_id, day), records in groupby(
        log_records, lambda r: (r.user_id, r.datetime.date())
    ):
        employees.append(_build_record(user_id, day, tuple(records)))

    result = Root()
    result.employees = employees
    return result


def get_authentications_info(date_from, date_to, **kwargs):
    """Возвращает данные о количестве аутентификаций через форму входа в ЭДО.

    :param date_from: Дата начала периода, за который нужно загрузить
        информацию.
    :type date_from: datetime.date

    :param date_to: Дата окончания периода, за который нужно загрузить
        информацию.
    :type date_to: datetime.date

    :rtype: :class:`edu_to_analytics.service.types.Root`
    """
    return _get_data(date_from, date_to, AUTH_LOGIN)


def get_esia_authentications_info(date_from, date_to, **kwargs):
    """Возвращает данные о количестве аутентификаций через ЕСИА.

    :param date_from: Дата начала периода, за который нужно загрузить
        информацию.
    :type date_from: datetime.date

    :param date_to: Дата окончания периода, за который нужно загрузить
        информацию.
    :type date_to: datetime.date

    :rtype: :class:`edu_to_analytics.service.types.Root`
    """
    return _get_data(date_from, date_to, AUTH_LOGIN_ESIA)


class LessonsStatsProvider(object):

    """Провайдер данных показателя LessonsStat."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @cached_property
    def synced_units(self):
        """Синхронизируемые учреждения."""
        unit_ct = ContentType.objects.get_for_model(Unit)
        dont_sync_unit_ids = SyncObject.objects.filter(
            content_type=unit_ct, dont_sync=True
        ).values_list('object_id')
        synced_units = Unit.objects.exclude(pk__in=dont_sync_unit_ids)
        return synced_units

    @cached_property
    def edo_units(self):
        """Образовательные учреждения."""
        return self.synced_units.filter(type_id=UnitType.SCHOOL)

    @cached_property
    def ministries(self):
        """Министерства учреждений."""
        return {unit: unit.get_root() for unit in self.synced_units}

    @cached_property
    def administrations(self):
        """Управления учреждений."""
        administrations = {}
        for unit in self.edo_units:
            query = unit.get_ancestors().filter(
                type_id=UnitType.ADMINISTRATION
            )
            if query:
                administrations[unit] = query[0]
        return administrations

    @cached_property
    def directions(self):
        """Направления в учреждениях."""
        directions = DirectionModel.objects.filter(
            unit__in=self.edo_units
        ).select_related('unit')
        return directions

    @cached_property
    def period(self):
        """Период на переданную дату."""
        return Period.objects.filter(
            Period.get_date_in_intervals_filter(self.kwargs['date_to'])
        ).first()

    @cached_property
    def groups(self):
        """Группы в учреждениях."""
        return GroupModel.objects.filter(
            unit__in=self.synced_units,
            period=self.period
        )

    @cached_property
    def group_count(self):
        """Количество групп по направлениям."""
        group_count = defaultdict(int)

        query = self.groups.order_by(
            'department__direction'
        ).values_list(
            'department__direction'
        ).annotate(Count('department__direction'))

        for direction_id, group_cnt in query:
            group_count[direction_id] = group_cnt

        return group_count

    @cached_property
    def pupils(self):
        """Учащиеся в группах."""
        return PupilModel.objects.filter(
            PupilModel.get_date_in_intervals_filter(self.kwargs['date_to']),
            group__in=self.groups
        )

    @cached_property
    def pupil_count(self):
        """Количество учащихся по направлениям."""
        pupil_count = defaultdict(int)

        query = self.pupils.order_by(
            'group__department__direction'
        ).values_list(
            'group__department__direction'
        ).annotate(Count('group__department__direction'))

        for direction_id, pupil_cnt in query:
            pupil_count[direction_id] = pupil_cnt

        return pupil_count

    @cached_property
    def lessons(self):
        """Занятия в группах учреждений."""
        lessons = LessonModel.objects.filter(
            lessongroup__group__unit__in=self.synced_units,
            date__range=(self.kwargs['date_from'], self.kwargs['date_to'])
        ).select_related(
            'lessongroup__group__unit__primary_direction',
            'employee__person', 'subject', 'lessongroup__subgroup'
        )
        return lessons

    @cached_property
    def direction_lessons(self):
        """Наличие расписание по направлениям."""
        result = defaultdict(bool)

        query = self.lessons.values_list(
            'lessongroup__group__department__direction'
        ).annotate(Count('id'))

        for direction_id, lesson_count in query:
            result[direction_id] = bool(lesson_count)

        return result

    @cached_property
    def group_subject_lessons_count(self):
        """Количество занятий в группе по предмету."""
        result = defaultdict(int)

        query = self.lessons.values_list(
            'lessongroup__group', 'subject'
        ).annotate(Count('id'))

        for group_id, subj_id, cnt in query:
            result[(group_id, subj_id)] = cnt

        return result

    @cached_property
    def lesson_homeworks(self):
        """Наличие задания на занятие."""
        result = defaultdict(bool)
        for lesson in self.lessons:
            result[lesson.id] = bool(
                NextLessonHomework.get_for_taget_lesson(lesson)
            )
        return result

    @cached_property
    def marks(self):
        """Оценки."""
        marks = MarkModel.objects.filter(
            task__lesson__in=self.lessons,
            pupil__in=self.pupils
        ).select_related(
            'task__lesson', 'pupil__person'
        )
        return marks

    @cached_property
    def lesson_marks(self):
        """Оценки учащихся по занятиям."""
        lesson_marks = defaultdict(list)

        for mark in self.marks:
            lesson_marks[mark.task.lesson_id].append(
                (
                    mark.pupil_id,
                    mark.pupil.person.fullname,
                    mark.task.lesson.date
                )
            )
        return lesson_marks

    @cached_property
    def lesson_marks_counts(self):
        """Количество оценок по урокам по весу."""
        result = defaultdict(lambda: defaultdict(int))

        query = self.marks.values_list(
            'task__lesson', 'markvalue', 'marksuffix'
        ).annotate(Count('id'))

        for lesson_id, markvalue_id, marksuffix_id, mark_count in query:
            normalized = normalize_mark(markvalue_id, marksuffix_id)
            result[lesson_id][normalized] = mark_count

        return result

    @cached_property
    def attendance_count(self):
        """Кол-во отсутствующих или болеющих по занятиям."""
        return dict(
            Attendance.objects
            .filter(pupil__in=self.pupils, lesson__in=self.lessons)
            .order_by('lesson').values_list('lesson')
            .annotate(Count('lesson'))
        )

    @cached_property
    def group_pupil_count(self):
        """Кол-во учащихся в группах."""
        return dict(
            self.pupils.order_by('group')
            .values_list('group').annotate(Count('group'))
        )

    @cached_property
    def subgroup_pupil_count(self):
        """Кол-во учащихся в подгруппах."""
        return dict(
            SubGroupPupil.objects
            .filter(pupil__in=self.pupils, subgroup__group__in=self.groups)
            .order_by('subgroup').values_list('subgroup')
            .annotate(Count('subgroup'))
        )

    @cached_property
    def lesson_pupil_count(self):
        """Количество присутствующих на занятии."""
        pupil_count = defaultdict(int)

        # Кол-во в группе или подгруппе - болеющие = итого
        for lesson in self.lessons:
            if lesson.lessongroup.subgroup is not None:
                full_count = (
                    self.subgroup_pupil_count.get(
                        lesson.lessongroup.subgroup_id,
                        0
                    )
                )
            else:
                full_count = self.group_pupil_count.get(
                    lesson.lessongroup.group_id,
                    0
                )

            pupil_cnt = full_count - self.attendance_count.get(lesson.id, 0)

            pupil_count[lesson.id] = pupil_cnt

        return pupil_count


class DirectionIterator(object):

    """Итератор по направлениям."""

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        for direction in self.data.directions:
            item = Direction()

            item.id = direction.id
            item.name = direction.name

            ministry = self.data.ministries[direction.unit]

            item.ministry_id = ministry.id
            item.ministry_full_name = ministry.full_name

            administration = self.data.administrations.get(direction.unit)
            if administration is not None:
                # Могут быть ОДО, подчиненные министерству
                item.administration_id = administration.id
                item.administration_full_name = administration.full_name

            item.unit_id = direction.unit_id
            item.unit_full_name = direction.unit.full_name
            item.unit_short_name = direction.unit.short_name

            item.group_count = self.data.group_count[direction.id]
            item.pupil_count = self.data.pupil_count[direction.id]

            item.has_lessons = self.data.direction_lessons[direction.id]

            yield item


class LessonIterator(object):

    """Итератор по занятиям."""

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        for lesson in self.data.lessons:
            item = Lesson()

            item.lesson_id = lesson.id
            item.lesson_date = lesson.date
            item.lesson_held = lesson.was_held

            item.unit_id = lesson.lessongroup.group.unit_id
            item.unit_full_name = lesson.lessongroup.group.unit.full_name
            item.unit_short_name = lesson.lessongroup.group.unit.short_name

            direction_id = lesson.lessongroup.group.unit.primary_direction_id

            item.direction_id = direction_id
            item.direction_name = (
                lesson.lessongroup.group.unit.primary_direction.name
            )

            group_id = lesson.lessongroup.group_id

            item.group_id = group_id

            item.teacher_id = lesson.employee.id
            item.teacher_name = lesson.employee.person.fullname

            subject_id = lesson.subject_id

            item.subject_id = subject_id
            item.subject_name = lesson.subject.name

            item.hours_count = (
                self.data.group_subject_lessons_count[(group_id, subject_id)]
            )

            item.homework = self.data.lesson_homeworks[lesson.id]

            item.marks = [
                mark for mark in MarkIterator(
                    self.data.lesson_marks[lesson.id]
                )
            ]

            item.pupils_visited = self.data.lesson_pupil_count[lesson.id]

            attr_name_template = 'mark_{}_count'
            for weight in range(2, 6):
                setattr(
                    item, attr_name_template.format(weight),
                    self.data.lesson_marks_counts[lesson.id][weight]
                )

            yield item


class MarkIterator(object):

    """Итератор по оценкам занятия."""

    def __init__(self, lesson_marks):
        self.lesson_marks = lesson_marks

    def __iter__(self):
        for pupil_id, pupil_fio, mark_date in self.lesson_marks:
            item = Mark()

            item.pupil_id = pupil_id
            item.pupil_fio = pupil_fio
            item.date = mark_date

            yield item


def get_lessons_stat(**kwargs):
    """Статистика по направлениям и занятиям."""
    data = LessonsStatsProvider(**kwargs)

    result = UnitsRoot()
    result.directions = [direction for direction in DirectionIterator(data)]
    result.lessons = [lesson for lesson in LessonIterator(data)]

    return result


CODE_METHOD_MAP = {
    'LoginCounter': get_authentications_info,
    'ESIALoginCounter': get_esia_authentications_info,
    'LessonsStat': get_lessons_stat,
}


def get_content(out_format=None, **kwargs):
    """Получает и сериализует данные в указанный формат."""

    out_format = out_format or 'xml'

    root = CODE_METHOD_MAP[kwargs['report_code']](**kwargs)

    if out_format == 'xml':
        result = tostring(
            get_object_as_xml(root), pretty_print=True
        )
    else:
        raise NotImplementedError(
            '{} format is not implemented'.format(out_format)
        )

    return result


class _UnitListenerBase(object):

    """Определяет метод получения записи синхронизации учреждения."""

    @staticmethod
    def get_unit_sync(unit_id):
        if not Unit.objects.filter(pk=unit_id).exists():
            raise ApplicationLogicException('Организация не найдена.')

        sync, _ = SyncObject.objects.get_or_create(
            content_type=ContentType.objects.get_for_model(Unit),
            object_id=unit_id
        )
        return sync


class UnitWindowListener(_UnitListenerBase):

    """Добавляет чекбокс в окно редактирования учреждения."""

    listen = (
        'extedu.unit.main/Pack/ObjectEditWindowAction',
    )

    def after(self, request, context, response):
        win = response.data

        sync = self.get_unit_sync(
            get_id_value(context, 'extedu.unit.main.actions.Pack')
        )

        analytics_sync_checkbox = ExtCheckBox(
            label='Не отправлять в систему "Аналитика"',
            name='analytics_dont_sync',
            checked=sync.dont_sync
        )

        main_tab = win._tab_container.items[0]
        main_tab.items.append(analytics_sync_checkbox)


class UnitSaveListener(_UnitListenerBase):

    u"""Проставляет флаг синхронизации при успешном сохранении учреждения."""

    listen = (
        'extedu.unit.main/Pack/SaveAction',
    )

    def after(self, request, context, response):
        if response.data.get('success', False):
            dont_sync = extract_bool(request, 'analytics_dont_sync')
            data = response.data.get('data', False)
            unit_id = data.get('id', 0)
            sync = self.get_unit_sync(unit_id)
            sync.dont_sync = dont_sync
            sync.clean_and_save()
